//Compile:/C

//!Text+ OGS_IniSys()   Class to create '.INI'-files

// Class   OGS_IniSys()       The class function
// 
// IniNew([f])                Create a INI-object for file 'f' or 'PROGNAME'
// IniGet([f],s,p,d)          See: Class Method Get()
// 
// Var     Changed            TRUE if file contents is changed      ReadOnly
// Var     Closed             TRUE if the file is closed            ReadOnly
// Var     FileName           Name of the '.INI'-file               ReadOnly
// Var     Locked             TRUE if the file is locked            ReadOnly
// 
// Method  Close()            Close the file
// Method  DelParameter(s,p)  Delete parameter 'p'
// Method  DelSection(s)      Delete section 's'
// Cls Mth Get([f],s,p,d)     Get value from 'p' in section 's'
//                            or default 'd' from file 'f'
// Method  Get(s,p,d)         Get value from 'p' in section 's'
//                            or default 'd'
// Method  Parameters(s)      Get list with parameters for section 's'
// Method  Put(s,p,v)         Put value 'v' in parameter 'p'
//                            for section 's'
// Method  Sections()         Get a list of sections
// Method  Lock()             Lock the file
// Method  Update(l)          Update the file and unlock if 'l'==TRUE

// // Ŀ
// //  Example using the 'OGS_IniSys()'-functions/methods and variables     
// // 
// #Include "OGSLIB.CH"
// //  OGSLIB.CH contains Ŀ
// //   #xTranslate IniNew([<f>]) => OGS_IniSys():New(...)                  
// // 
// Procedure Main()
// Local IniObj:=IniNew(), AllSections, AllParameters
//    DspStat(IniObj)
//    IniObj:Lock()
//    DspStat(IniObj)
//    IniObj:Update(TRUE)
//    DspStat(IniObj)
//    IniObj:Put("MainSect","MainPar1","MainParValue1")
//    IniObj:Put("MainSect","MainPar2","MainParValue2New")
//    IniObj:Put("MainSect","MainPar3",{1,{2,{3,4},5}})
//    IniObj:Put("SubsSect","SubsPar1","SubsParValue1")
//    IniObj:Put("MainSect","MainPar3","MainParValue3")
//    IniObj:Put("MainSect","MainPar2","MainParValue2New")
//    IniObj:Put("SubsSect","SubsPar2","SubsParValue2New")
//    IniObj:Put("SubsSect","SubsPar3","SubsParValue3")
//    DspStat(IniObj)
//    ? IniObj:Get("MainSect","MainPar1")
//    ? IniObj:Get("MainSect","MainPar2")
//    ? "MainPar3 :=",IniObj:Get("MainSect","MainPar3")[2][2][2]
//    ? IniObj:Get("SubsSect","SubsPar1")
//    ? IniObj:Get("MainSect","MainPar3")
//    ? IniObj:Get("SubsSect","SubsPar2")
//    ? IniObj:Get("SubsSect","SubsPar3")
//    ? IniObj:Get("TestSect","TestPar1",321)
//    ? IniObj:Get("TestSect","TestPar2",CToD("01/01/93"))
//    ? IniObj:Get("TestSect","TestPar3",TRUE)
//    ? IniObj:Get("TestSect","TestPar4",FALSE)
//    DspStat(IniObj)
//    AllSections:=IniObj:Sections()
//    DspStat(IniObj)
//    IniObj:DelSection("SubsSect")
//    AllSections:=IniObj:Sections()
//    AllParameters:=IniObj:Parameters("TestSect")
//    IniObj:DelParameter("TestSect","TestPar3")
//    AllParameters:=IniObj:Parameters("TestSect")
//    IniObj:Close()
//    DspStat(IniObj)
// Return
//
// Procedure DspStat(IniObj)
//    ? "Changed:",IniObj:Changed,"  Locked:";
//      ,IniObj:Locked,"  Open:",!IniObj:Closed
// Return

//!Text-

#Include "ERROR.CH"
#Include "FILEIO.CH"
#Include "FUNCKY.CH"
#Include "INIOBJ.CH"
#Include "CLASS(Y).CH"

Static IniList:={}

// ͻ
//  Request a new 'INI'-object                                              
// Ķ
//  InpFileReChar    Name of the file to open                            
//           Other   The current programname is used for the filename.   
//                   If the filename does not contain an drive and/or    
//                   path indication, Then the '.INI'-file will be opened
//                   in the same directory as the '.EXE'-file.           
// Ķ
//  RetSelf  Object  The object it'self.                                 
// ͼ
Function OGS_IniNew(FileRequested)
Local RetVal,FileName,FileExt,FileHandle,FileNamExt,Index
   FileName:=If(NotC(FileRequested) OR Empty(FileRequested);
               ,ComlProgram()                              ;
               ,Upper(AllTrim(FileRequested))              )
   If Empty(FileExt:=SubStr(FParseFx(FileName),2))
      FileExt:="INI"
   Else
      FileName:=SubStr(FileName,1,Len(FileName)-(Len(FileExt)+1))
   End
   If Empty(FParseD(FileName)+FParseDr(FileName))
      FileName:=Upper(ComlDrive()+":"+ComlPath()+FileName)
   End
   If !File(FileName+"."+FileExt)
      FileHandle:=FCreate(FileName+"."+FileExt,FC_NORMAL)
      If FError()<>0 Then ForceError(EG_CREATE,FileHandle,FileName+"."+FileExt)
      FClose(FileHandle)
   End
   FileNamExt:=FileName+"."+FileExt
   If (Index:=aScan(IniList,{|x|x[1]==FileNamExt}))<>0
      RetVal:=IniList[Index][2]
      RetVal:OpenCount++
   Else
      RetVal:=OGS_IniSys():New(FileName,FileExt)
   End
Return RetVal

// Ŀ
//  Now define the 'Ini'-class and it's methods                             
// 
Create Class OGS_IniSys
Export: // XXX //
   Var           BChange               // TRUE if buffer contents is changed
   Var           CurrSection           // Current section in buffer
   Var           FileHandle            // Handle for the file
   Var           SectData              // Data for current section
   Var           SemaHandle            // Handle used for Novell semaphore
// XXX // Export:
   Var           Changed      // XXX // ReadOnly // TRUE if file contents is changed
   Var           Closed       // XXX // ReadOnly // TRUE if the file is closed
   Var           FileName     // XXX // ReadOnly // Name of the open file
   Var           FileExt      // XXX // ReadOnly // Extention of the open file
   Var           Locked       // XXX // ReadOnly // TRUE if the file is locked
   Var           Open         // XXX // ReadOnly // TRUE if a file is open
   Var           OpenCount             // Counts number of instances
// XXX // Hidden:
   Message       DoFlush      Method IniDoFlush
   Message       DoInsert     Method IniDoInsert
   Message       DoLock       Method IniDoLock
   Message       GetSection   Method IniGetSection
   Message       LockError    Method IniLockError
// XXX // Export:
   Method                            Init
   Class Method                      InitClass
   Message       Close        Method IniClose
   Message       DelParameter Method IniDelParameter
   Message       DelSection   Method IniDelSection
   Class Message Get          Method IniClsGet
   Message       Get          Method IniDoGet
   Message       Lock         Method IniSetLock
   Message       Parameters   Method IniParameters
   Message       Put          Method IniPut
   Message       Sections     Method IniSections
   Message       Update       Method IniUpdate
EndClass

// ͻ
//  Initialize the class-object                                             
// Ķ
//  RetSelf  Object  The object it'self.                                 
// ͼ
Method Function InitClass(),()
Return Self

// ͻ
//  Open an '.INI'-file                                                     
// Ķ
//  InpFileNaChar    Name of the file to open.                           
// Ķ
//  RetSelf  Object  The object it'self.                                 
// ͼ
Method Function Init(FileName,FileExt),()
   ::FileHandle:=FOpen(FileName+"."+FileExt,O_READWRITE+S_DENYNONE)
   If FError()<>0 Then ForceError(EG_OPEN,NIL,FileName+"."+FileExt)
   ::Changed:=FALSE
   ::BChange:=FALSE
   ::Locked:=FALSE
   ::SectData:={}
   ::FileName:=FileName
   ::FileExt:=FileExt
   ::Open:=TRUE
   ::Closed:=FALSE
   ::CurrSection:=""
   ::SemaHandle:=SemOpen(TrueName(FileName+"."+FileExt))
   ::OpenCount:=1
   aAdd(IniList,{FileName+"."+FileExt,Self})
Return Self

// ͻ
//  Close an open '.INI'-file                                               
// Ķ
//  RetSelf  Object  The object it'self.                                 
// ͼ
Method Function IniClose()
Local FileNamExt,Index
   If !::Closed
      @:IniDoFlush(TRUE,FALSE)
      ::OpenCount--
      If ::OpenCount<1
         If ::SemaHandle<>0 Then SemClose(::SemaHandle)
         FClose(::FileHandle)
         ::Open:=FALSE
         ::Closed:=TRUE
         FileNamExt:=::FileName+"."+::FileExt
         If (Index:=aScan(IniList,{|x|x[1]==FileNamExt}))<>0
            aDel(IniList,Index)
            aSize(IniList,Len(IniList)-1)
         End
      End
   End
Return Self

// ͻ
//  Open/Read/Close an INI-object                                           
// Ķ
//  InpFileNaChar    Name of the INI-file to use.                        
//     SectioChar    Name of the section where the parameter resides.    
//     Parm  Char    Name of the parameter.                              
//     DefaulAny     Value to return if parameter is not found.          
// Ķ
//  RetRetValAny     Value of the 'Parameter' from requested 'Section'   
// ͼ
Method Function IniClsGet(FileName,Section,Parameter,Default)
Local RetVal,IniObj:=IniNew(FileName)
   RetVal:=IniObj:Get(Section,Parameter,Default)
   IniObj:Close()
Return RetVal

// ͻ
//  Create an error object for an locking error                             
// Ķ
//  InpSCode Numeric Sub-code for error object.                          
// Ķ
//  RetIniErrObject  Error object.                                       
// ͼ
Method Function IniLockError(SCode)
Local IniError:=ErrorNew()
   IniError:CanRetry:=TRUE
   IniError:Gencode:=EG_LOCK
   IniError:Severity:=ES_ERROR
   IniError:SubSystem:=CompModule
   IniError:SubCode:=SCode
   IniError:FileName:=::FileName+"."+::FileExt
   IniError:Description:="Can not set lock on file"
Return IniError

// ͻ
//  Delete 'Section' from the file.                                         
// Ķ
//  InpSectioChar    Name of the section to delete.                      
// Ķ
//  Ret      Logical TRUE if success ( Note: The file will be locked )   
// ͼ
Method Function IniDelSection(Section)
Local IniError,Retry:=TRUE,RetVal:=FALSE
   If IsC(Section) AND !Empty(Section) AND !::Closed
      While Retry
         Retry:=FALSE
         If !@:IniDoLock()
            If IniError==NIL Then IniError:=@:IniLockError(1)
            If NotL(Retry:=Eval(ErrorBlock(),IniError)) Then Retry:=FALSE
         End
      End
      If RetVal:=::Locked
         ::CurrSection:="["+AllTrim(Section)+"]"
         ::Changed:=::BChange:=TRUE
         @:IniDoFlush(FALSE,TRUE)
      End
   End
Return RetVal

// ͻ
//  Delete an 'Parameter' that is in 'Section'.                             
// Ķ
//  InpSectioChar    Name of the section where the parameter resides.    
//     Parm  Char    Name of the parameter to delete.                    
// Ķ
//  Ret      Logical TRUE if success ( Note: The file will be locked )   
// ͼ
Method Function IniDelParameter(Section,Parm)
Local LocateStr,LocateLen,Index,RetVal:=FALSE,Retry:=TRUE,IniError
   If IsC(Section) AND !Empty(Section) AND !::Closed
      While Retry
         Retry:=FALSE
         If !@:IniDoLock()
            If IniError==NIL Then IniError:=@:IniLockError(2)
            If NotL(Retry:=Eval(ErrorBlock(),IniError)) Then Retry:=FALSE
         End
      End
      If RetVal:=::Locked
         @:IniGetSection(Section)
         LocateStr:=AllTrim(Upper(Parm))+"="
         LocateLen:=Len(LocateStr)
         For Index:=1 To Len(::SectData)
            If Upper(Left(::SectData[Index],LocateLen))==LocateStr
               While Index<Len(::SectData)
                  ::SectData[Index]:=::SectData[Index+1]
                  Index++
               End
               aSize(::SectData,Len(::SectData)-1)
               ::Changed:=::BChange:=TRUE
            End
         Next
      End
   End
Return RetVal

// ͻ
//  Write the data from internal buffer back to the file                    
// Ķ
//  RetSelf  Object  The object it'self.                                 
// ͼ
Method Function IniDoInsert()
Local Index
   FWriteLine(::FileHandle,CrLf+::CurrSection)
   If FError()==0
      For Index:=1 To Len(::SectData)
         FWriteLine(::FileHandle,::SectData[Index])
         If FError()<>0 Then Exit
      Next
   End
   If FError()<>0 Then ForceError(EG_WRITE,NIL,::FileName+"."+::FileExt)
Return Self

// ͻ
//  Flush and optional unlock an '.INI'-file.                               
// Ķ
//  InpUnLockLogical TRUE if the file must be unlocked.                  
//     DeleteLogical TRUE if the current section must not be saved.      
// Ķ
//  RetSelf  Object  The object it'self.                                 
// ͼ
Method Function IniDoFlush(UnLock,Delete)
Local NewHandle,IniLine,Inserted:=Delete,CopyThis:=TRUE
   If ::BChange
      NewHandle:=FCreate(::FileName+".BAK",0)
      FSeek(::FileHandle,0,0)
      While !FEof(::FileHandle)
         IniLine:=FReadLine(::FileHandle)
         If FError()==0;
            Then FWriteLine(NewHandle,IniLine);
            Else ForceError(EG_READ,NewHandle,::FileName+".BAK")
      End
      FSeek(NewHandle,0,0)
      FSeek(::FileHandle,0,0)
      While !FEof(NewHandle) AND FError()==0
         If !Empty(IniLine:=AllTrim(FReadLine(NewHandle)))
            If Left(IniLine,1)=="[" AND Right(IniLine,1)=="]"
               CopyThis:=TRUE
               If Upper(IniLine)==Upper(::CurrSection)
                  If !Inserted Then @:IniDoInsert()
                  Inserted:=TRUE
                  CopyThis:=FALSE
               Else
                  FWriteLine(::FileHandle,"")
               End
            End
            If CopyThis AND FError()==0
               FWriteLine(::FileHandle,IniLine)
            End
         End
      End
      If FError()==0
         If !Inserted Then @:IniDoInsert()
         If FError()==0 Then FClose(NewHandle)
         If FError()==0 Then FTruncate(::FileHandle)
         If FError()==0 Then FCommit(::FileHandle)
      End
      If FError()<>0 Then ForceError(EG_WRITE,NewHandle,::FileName+".???")
      ::BChange:=FALSE
   End
   If UnLock AND ::Locked
      ::Locked:=FALSE
      If ::SemaHandle<>0 Then SemSignal(::SemaHandle)
   End
Return Self

// ͻ
//  Get the value of an 'Parameter' that is in 'Section'.                   
// Ķ
//  InpSectioChar    Name of the section where the parameter resides.    
//     Parm  Char    Name of the parameter.                              
//     DefaulAny     Value to return if parameter is not found.          
// Ķ
//  RetRetValAny     Value of the 'Parameter' from requested 'Section'   
// ͼ
Method Function IniDoGet(Section,Parm,Default)
Local Index,LocateStr,LocateLen,RetVal,LenRet
   If IsC(Section) AND !Empty(Section) AND !::Closed
      @:IniGetSection(Section)
      LocateStr:=AllTrim(Upper(Parm))+"="
      LocateLen:=Len(LocateStr)
      For Index:=1 To Len(::SectData)
         If Upper(Left(::SectData[Index],LocateLen))==LocateStr
            RetVal:=StrToAny(SubStr(::SectData[Index],LocateLen+1))
            Exit
         End
      Next
   End
   If IsA(Default)
      If IsA(RetVal)
         If !IniArrComp(Default,RetVal) Then RetVal:=NIL
      Else
         RetVal:=NIL
      End
   End
   If ValType(RetVal)<>ValType(Default) AND Default<>NIL
      @:IniPut(Section,Parm,Default)
      RetVal:=Default
   End
Return RetVal
//
Static Function IniArrComp(Arr1,Arr2)
Local Len1:=Len(Arr1),Len2:=Len(Arr2),Typ1
   If Len1<>Len2 Then Return FALSE
   For Len1:=1 To Len2
      If (Typ1:=ValType(Arr1[Len1]))<>ValType(Arr2[Len1]) Then Return FALSE
      If Typ1=="A" AND !IniArrComp(Arr1[Len1],Arr2[Len1]) Then Return FALSE
   Next
Return TRUE

// ͻ
//  Read the requested section from '.INI'-file to '::SectData'             
// Ķ
//  InpSectioChar    Section to read the data from.                      
// Ķ
//  RetSelf  Object  The object it'self.                                 
// ͼ
Method Function IniGetSection(Section)
Local IniLine,UpSection:="["+Upper(AllTrim(Section))+"]"
   If !(UpSection==Upper(::CurrSection))
      @:IniDoFlush(FALSE,FALSE)
      ::SectData:={}
      ::CurrSection:="["+AllTrim(Section)+"]"
      FSeek(::FileHandle,0,0)
      While !FEof(::FileHandle)
         If Upper(AllTrim(FReadLine(::FileHandle)))==UpSection Then Exit
      End
      While !FEof(::FileHandle)
         If !Empty(IniLine:=AllTrim(FReadLine(::FileHandle)))
            If Left(IniLine,1)=="[" AND Right(IniLine,1)=="]" Then Exit
            aAdd(::SectData,IniLine)
         End
      End
   End
Return Self

// ͻ
//  Lock the file.                                                          
// Ķ
//  RetRetValLogical TRUE if the requested file is locked.               
// ͼ
Method Function IniSetLock()
   If !::Closed Then @:IniDoLock()
Return ::Locked

// ͻ
//  Get all parameters from 'Section'.                                      
// Ķ
//  InpSectioChar    Name of the section where the parameter resides.    
// Ķ
//  RetParaLsArray   Array with all parameters from specified section.   
// ͼ
Method Function IniParameters(Section)
Local ParametersList:={}
   If IsC(Section) AND !Empty(Section) AND !::Closed
      @:IniGetSection(Section)
      ParametersList:=aClone(::SectData)
   End
Return ParametersList

// ͻ
//  Save 'Value' for parameter 'Parm' in section 'Section'.                 
// Ķ
//  InpSectioChar    Name of the section where the parameter resides.    
//     Parm  Char    Name of the parameter.                              
//     Value Any     Value to put in 'Section' for 'Parameter'           
// Ķ
//  Ret      Logical TRUE if success ( Note: The file can be locked )    
// ͼ
Method Function IniPut(Section,Parm,Value)
Local Index,LocateStr,LocateLen,DatTyp,DatStr,RetVal:=FALSE,Retry:=TRUE,IniError
   If IsC(Section) AND !Empty(Section) AND !::Closed
      @:IniGetSection(Section)
      LocateStr:=AllTrim(Upper(Parm))+"="
      LocateLen:=Len(LocateStr)
      DatStr:=AnyToStr(Value)
      Index:=aScan(::SectData,{|x|Upper(Left(x,LocateLen))==LocateStr})
      If Index==0 OR !(::SectData[Index]==AllTrim(Parm)+"="+DatStr)
         If !::Locked
            While Retry
               Retry:=FALSE
               If !@:IniDoLock()
                  If IniError==NIL Then IniError:=@:IniLockError(3)
                  If NotL(Retry:=Eval(ErrorBlock(),IniError)) Then Retry:=FALSE
               End
            End
            If RetVal:=::Locked
               ::CurrSection:=""
               ::SectData:={}
               @:IniGetSection(Section)
               Index:=aScan(::SectData,{|x|Upper(Left(x,LocateLen))==LocateStr})
            End
         Else
            RetVal:=TRUE
         End
         If RetVal
            If Index==0
               aAdd(::SectData,AllTrim(Parm)+"=")
               Index:=Len(::SectData)
            End
            If !(::SectData[Index]==AllTrim(Parm)+"="+DatStr)
               ::SectData[Index]:=AllTrim(Parm)+"="+DatStr
               ::Changed:=::BChange:=TRUE
            End
         End
      Else
         RetVal:=TRUE
      End
   End
Return RetVal

// ͻ
//  Get all sections from the file.                                         
// Ķ
//  RetSectLsArray   Array with all sections listed.                     
// ͼ
Method Function IniSections()
Local SectionsList:={},IniLine
   If !::Closed
      @:IniDoFlush(FALSE,FALSE)
      FSeek(::FileHandle,0,0)
      While !FEof(::FileHandle)
         If !Empty(IniLine:=AllTrim(FReadLine(::FileHandle)))
            If Left(IniLine,1)=="[" AND Right(IniLine,1)=="]"
               IniLine:=SubStr(IniLine,2,Len(IniLine)-2)
               If !Empty(IniLine) Then aAdd(SectionsList,IniLine)
            End
         End
      End
   End
Return SectionsList

// ͻ
//  Lock the file.                                                          
// Ķ
//  Ret      Logical TRUE if the lock is set.                            
// ͼ
Method Function IniDoLock()
Local RetryCount:=0
   If !::Locked AND ::SemaHandle>0
      @:IniDoFlush(FALSE,FALSE)
      While !SemWait(::SemaHandle)
         If RetryCount>100 Then Exit
         Delay(10)
         RetryCount++
      End
   End
   If RetryCount<101 Then ::Locked:=TRUE
Return ::Locked

// ͻ
//  Write changed data back to the '.INI'-file, And optional release lock.  
// Ķ
//  InpUnlockLogical TRUE if the file must be unlocked.                  
//           Other   File will not be unlocked.                          
// Ķ
//  RetSelf  Object  The object it'self.                                 
// ͼ
Method Function IniUpdate(UnLock)
   If !::Closed Then @:IniDoFlush(If(IsL(UnLock),UnLock,FALSE),FALSE)
Return Self

// ͻ
//  Convert an array to a string                                            
// Ķ
//  InpInpArrArray   Array to convert                                    
// Ķ
//  RetOutStrChar    Converted array                                     
// ͼ
Function ArrToStr(InpArr)
Return "a'{"+DoArrToStr(InpArr)+"}'"
//
Static Function DoArrToStr(InpArr)
Local OutStr:="",ArrIndex,LenInpArr:=Len(InpArr),DatTyp
   If LenInpArr>0
      If IsA(InpArr[1])
         OutStr+="{"+DoArrToStr(InpArr[1])+"}"
      Else
         OutStr+=LTrim(Str(InpArr[1]))
      End
      For ArrIndex:=2 To LenInpArr
         DatTyp:=ValType(InpArr[ArrIndex])
         If DatTyp=="N"
            OutStr+=","+LTrim(Str(InpArr[ArrIndex]))
         ElseIf DatTyp=="A"
            OutStr+=",{"+DoArrToStr(InpArr[ArrIndex])+"}"
         Else
            OutStr+=",0"
         End
      Next
   End
Return OutStr

// ͻ
//  Convert an string to a array                                            
// Ķ
//  InpInpStrChar    String to convert with the leading '{' removed      
//     StrIndNumeric Current conversion position pointer                 
//     LenInpNumeric Number of bytes to convert                          
// Ķ
//  RetOutArrArray   Converted string                                    
// ͼ
Function StrToArr(InpStr)
Return DoStrToArr(SubStr(InpStr,4),0,Len(InpStr)-4)
//
Static Function DoStrToArr(InpStr,StrIndex,LenInpStr)
Local NumBuffer:="",ThisChar,OutArr:={}
   While StrIndex<LenInpStr
      StrIndex++
      If (ThisChar:=SubStr(InpStr,StrIndex,1))=="{"
         aAdd(OutArr,DoStrToArr(@InpStr,@StrIndex,LenInpStr))
         StrIndex++
         If (ThisChar:=SubStr(InpStr,StrIndex,1))=="}" Then Exit
      ElseIf ThisChar=="}"
         aAdd(OutArr,Val(NumBuffer))
         Exit
      ElseIf ThisChar==","
         aAdd(OutArr,Val(NumBuffer))
         NumBuffer:=""
      Else
         NumBuffer+=ThisChar
      End
   End
Return OutArr

// ͻ
//  Convert any(Char,Date,Numeric,Logical,Array) to a string                
// Ķ
//  InpValue Any     Item to convert                                     
// Ķ
//  RetRetValChar    Item converted to string                            
// ͼ
Function AnyToStr(Value)
Local DatTyp:=ValType(Value),RetVal
   Do Case
   Case DatTyp=="C"
      RetVal:=AllTrim(Value)
   Case DatTyp=="D"
      RetVal:="d'"+DToC(Value)+"'"
   Case DatTyp=="L"
      RetVal:="l'"+If(Value,"TRUE","FALSE")+"'"
   Case DatTyp=="N"
      RetVal:="n'"+AllTrim(Str(Value))+"'"
   Case DatTyp=="A"
      RetVal:=ArrToStr(Value)
   Otherwise
      RetVal:="?'Invalid'"
   End
Return RetVal

// ͻ
//  Convert result from 'AnyToStr()' back to the original value             
// Ķ
//  InpStrDatChar    String to convert                                   
// Ķ
//  RetRetValItem    String converted to item                            
// ͼ
Function StrToAny(StrDat)
Local RetVal,DatTyp,RetDat,LenDat
   DatTyp:=Upper(Left(StrDat,2)+Right(StrDat,1))
   If Right(DatTyp,2)=='""' Then DatTyp:=Left(DatTyp,1)+"''"
   If Len(StrDat)>2 AND Right(DatTyp,2)=="''"
      RetDat:=SubStr(StrDat,3,Len(StrDat)-3)
      Do Case
      Case DatTyp=="C''"
         RetVal:=RetDat
      Case DatTyp=="D''"
         RetVal:=CToD(RetDat)
      Case DatTyp=="L''"
         RetDat:=Upper(RetDat)
         LenDat:=Len(RetDat)
         If LenDat>0
            RetVal:=(LenDat<5 AND RetDat==Left("TRUE",LenDat));
                 OR (LenDat<4 AND RetDat==Left("YES" ,LenDat));
                 OR (LenDat<3 AND RetDat==Left("JA"  ,LenDat))
         Else
            RetVal:=FALSE
         End
      Case DatTyp=="N''"
         RetVal:=Val(RetDat)
      Case DatTyp=="A''"
         RetVal:=StrToArr(StrDat)
      Otherwise
         RetVal:=NIL
      End
   Else
      RetVal:=StrDat
   End
Return RetVal

