/*Ŀ
** Clipper functions                                                        
**
*****************************************************************************/
#pragma  check_stack-

#include "dos.h"
#include "extend.api"
#include "item.api"
#include "vm.api"
#include "error.api"

#define EM_SEMOPEN   "SEMOPEN"
#define EM_DEVONLINE "DEVONLINE"
#define EM_SHADOW    "SHADOW"
#define EM_DOTFORMAT "DOTFORMAT"
#define EM_ATTR2REV  "ATTR2REV"
#define EM_ATTR2NOR  "ATTR2NOR"
#define EM_TRUENAME  "TRUENAME"
#define EM_SEMVALUE  "SEMVALUE"
#define EM_SEMCOUNT  "SEMCOUNT"
#define EM_SEMWAIT   "SEMWAIT"
#define EM_SEMSIGNAL "SEMSIGNAL"
#define EM_SEMCLOSE  "SEMCLOSE"

#define ES_ERROR     2

#define toupper(c)   ( (c)>='a' && (c)<='z' ? (c)&0xDF : (c) )
#define tolower(c)   ( (c)>='A' && (c)<='Z' ? (c)|0x20 : (c) )

extern USHORT _J_ConNumb(void);
extern USHORT _J_IsNovell(void);
extern USHORT _J_CurDrive(void);
extern USHORT _J_TrueName(BYTEP,BYTEP);
extern ULONG  _J_SemOpen(BYTEP,SHORT,SHORT);
extern SHORT  _J_SemVal(ULONG);
extern USHORT _J_SemOpc(ULONG);
extern SHORT  _J_SemFn(ULONG,USHORT,USHORT);
extern USHORT _J_IsProtMode(void);

/*Ŀ
** An error has happened                                                    
**
*****************************************************************************/
ERRCODE J_ErrorHandler(USHORT uiSubCode,BYTEP fpOperation)
{
    ERRORP pError;
    ERRCODE uiErrCode;
    pError=_errNew();
    _errPutDescription(pError,"OGS_ClpFn error");
    _errPutFileName(pError,"");
    _errPutFlags(pError,0);
    _errPutGenCode(pError,0);
    _errPutOperation(pError,fpOperation);
    _errPutOsCode(pError,0);
    _errPutSeverity(pError,ES_ERROR);
    _errPutSubCode(pError,uiSubCode);
    _errPutSubSystem(pError,"OGSCLPFN");
    _errPutTries(pError,0);
    uiErrCode=_errLaunch(pError);
    _errRelease(pError);
    _ret();
    return uiErrCode;
}

/*Ŀ
** logical:=J_TstDevOnLine(str,len) support for J_DEVONLIN()                
**
*****************************************************************************/
HIDE void J_TstDevOnLine(BYTEP sInpStr,USHORT nInpLen)
{
    BYTEP sTmpStr;
    BYTE DevName[9];
    USHORT nIndex;
    if(0) // (nInpLen>0)
    {
        sTmpStr=sInpStr;
        while(*sInpStr)
        {
            if(*sInpStr=='\\' || *sInpStr==':') sTmpStr=sInpStr+1;
            sInpStr++;
        }
        while(*sTmpStr==' ') sTmpStr++;
        nIndex=0;
        while(nIndex<9)
        {
            if(*sTmpStr && *sTmpStr!='.') DevName[nIndex]=toupper(*sTmpStr++);
            else DevName[nIndex]=' ';
            nIndex++;
        }
        DevName[8]=NIL;
        if(DevName[1]!=' ') // ?? [1]
        {
            /* Find 'DevName' in list                                */
            /* If it is a 'BlockDevice' then return NIL              */
            /* If the 'DevName' dous not exist then return TRUE      */
            /* If 'DevName'=='LPTx    ' then test status for 'x'     */
            /* ElseIf 'DevName'=='COMy    ' then test status for 'y' */
            /* Else return TRUE                                      */
            _retl(1);
        }
        else (void)J_ErrorHandler(1,EM_DEVONLINE);
    }
    else (void)J_ErrorHandler(2,EM_DEVONLINE);
}

/*Ŀ
** Logical:=J_DEVONLIN([DeviceName])                                        
** If ERROR then Result==NIL                                                
**                                                                          
** Return TRUE if the device is ready to receive data                       
**
*****************************************************************************/
CLIPPER J_DEVONLIN(void)
{
    if(PCOUNT==1 && ISCHAR(1)) J_TstDevOnLine(_parc(1),_parclen(1));
    else (void)J_ErrorHandler(3,EM_DEVONLINE);
}

/*Ŀ
** str:=SetShadow(str,len,attr) support for J_SHADOW()                      
**
*****************************************************************************/
HIDE void J_SetShadow(BYTEP sInpStr,USHORT nInpLen,BYTE nAttr)
{
    BYTEP sResult,sDst;
    USHORT nLoop;
    HANDLE xvHandle;
    if(nInpLen>0)
    {
        if(xvHandle=_xvalloc(nInpLen+1,0))
        {
            if((sDst=sResult=(BYTEP)_xvlock(xvHandle))!=NULL)
            {
                nLoop=nInpLen>>1;
                while(nLoop--)
                {
                    *sDst++=*sInpStr++;
                    *sDst++=nAttr;
                    sInpStr++;
                }
                *sDst=NIL;
                _retclen(sResult,nInpLen);
                _xvunlock(xvHandle);
            }
            else (void)J_ErrorHandler(1,EM_SHADOW);
            _xvfree(xvHandle);
        }
        else (void)J_ErrorHandler(2,EM_SHADOW);
    }
    else _retc("");
}

/*Ŀ
** String:=J_SHADOW(SavedScreen,Attribute)                                  
** If ERROR then Result==NIL                                                
**                                                                          
** Return input-string with attributes replaced by shadow attributes        
**
*****************************************************************************/
CLIPPER J_SHADOW(void)
{
    if(PCOUNT==2 && ISCHAR(1) && ISNUM(2))
      J_SetShadow(_parc(1),_parclen(1),_parni(2));
    else _retc("");
}

/*Ŀ
** String:=J_DOTFORMAT(String)                                              
** If ERROR then Result==NIL                                                
**                                                                          
** Return input-string formatted with dot's and a ','                       
**
*****************************************************************************/
CLIPPER J_DOTFORMA(void)
{
    BYTEP sResult,sDst,sSrc;
    BYTE ch;
    USHORT nLoop,nInpLen;
    HANDLE xvHandle;
    if(PCOUNT==1 && ISCHAR(1))
    {
        nInpLen=nLoop=_parclen(1);
        if(nInpLen>0)
        {
            if(xvHandle=_xvalloc(nInpLen+1,0))
            {
                if((sDst=sResult=(BYTEP)_xvlock(xvHandle))!=NULL)
                {
                    sSrc=_parc(1);
                    if(*sSrc==',')
                    {
                        nLoop--;
                        *sSrc++;
                        *sDst++=' ';
                    }
                    while(nLoop)
                    {
                        ch=*sSrc++;
                        if(ch=='.')
                        {
                            ch=',';
                        }
                        else
                        {
                            if(ch==',') ch='.';
                        }
                        *sDst++=ch;
                        nLoop--;
                    }
                    *sDst=NIL;
                    _retclen(sResult,nInpLen);
                    _xvunlock(xvHandle);
                }
                else (void)J_ErrorHandler(1,EM_DOTFORMAT);
                _xvfree(xvHandle);
            }
            else (void)J_ErrorHandler(2,EM_DOTFORMAT);
        }
        else _retc("");
    }
    else (void)J_ErrorHandler(3,EM_DOTFORMAT);
}

/*Ŀ
** Attribute:=J_ROLOC(Attribute)                                            
** If ERROR then Result==7                                                  
**                                                                          
** Replace the funcky 'ROLOC()'-function                                    
**
*****************************************************************************/
CLIPPER J_ROLOC(void)
{
    USHORT attr;
    if(PCOUNT==1 && ISNUM(1))
    {
        attr=_parni(1)<<4;
        _retni((((attr&0x0F00)>>8)|attr)&0x00FF);
    }
    else _retni(7);
}

/*Ŀ
** Conversion routine to support 'J_ATTR2NOR()' and 'J_ATTR2REV()'          
**
*****************************************************************************/
HIDE void J_AttrToAsc(USHORT at1,USHORT at2)
{
// ----------------------------------------------- Return CLIPPER-color values
//  BYTEP sDst,sSrc;
//  BYTE sResult[8];
//  static char *colors[]={"N" ,"B" ,"G" ,"BG" ,"R" ,"RB" ,"GR" ,"W"
//                        ,"N+","B+","G+","BG+","R+","RB+","GR+","W+"};
//  sDst=sResult;
//  sSrc=colors[at1];
//  while(*sSrc) *sDst++=*sSrc++;
//  *sDst++='/';
//  sSrc=colors[at2];
//  while(*sSrc) *sDst++=*sSrc++;
// --------------------------------------------- Return numerical color values
    BYTEP sDst;
    BYTE sResult[6];
    sDst=sResult;
    if(at1>9)
    {
        *sDst++=49;
        *sDst++=38+at1;
    }
    else *sDst++=48+at1;
    *sDst++='/';
    if(at2>9)
    {
        *sDst++=49;
        *sDst++=38+at2;
    }
    else *sDst++=48+at2;
// ---------------------------------------------- Common code for both methods
    *sDst=NIL;
    _retc(sResult);
}

/*Ŀ
** ColorString:=J_ATTR2REV(Attribute)                                       
** If ERROR then Result==NIL                                                
**                                                                          
** Replace the funcky 'ATTR2A()'-function                                   
**
*****************************************************************************/
CLIPPER J_ATTR2REV(void)
{
    USHORT attr;
    if(PCOUNT==1 && ISNUM(1))
    {
        attr=_parni(1);
        J_AttrToAsc((attr>>4)&0x0F,attr&0x0F);
    }
    else (void)J_ErrorHandler(1,EM_ATTR2REV);
}

/*Ŀ
** ColorString:=J_ATTR2NOR(Attribute)                                       
** If ERROR then Result==NIL                                                
**                                                                          
** Replace the funcky 'ATTR2A()'-function                                   
**
*****************************************************************************/
CLIPPER J_ATTR2NOR(void)
{
    USHORT attr;
    if(PCOUNT==1 && ISNUM(1))
    {
        attr=_parni(1);
        J_AttrToAsc(attr&0x0F,(attr>>4)&0x0F);
    }
    else (void)J_ErrorHandler(1,EM_ATTR2NOR);
}

/*Ŀ
** String:=J_ConNumb()                                                      
**                                                                          
** Return CONNECTION('xx') if Novell, Else return '00'                      
**
*****************************************************************************/
CLIPPER J_CONNUMB(void)
{
    USHORT cNum;
    BYTE sResult[3];
    cNum=_J_ConNumb();
    sResult[0]=cNum&0x00FF;
    sResult[1]=cNum>>8;
    sResult[2]=NIL;
    _retc(sResult);
}

/*Ŀ
** String:=J_CURDRIVE()                                                     
**                                                                          
** Replace the funcky 'CURDRIVE()'-function                                 
**
*****************************************************************************/
CLIPPER J_CURDRIVE(void)
{
    BYTE sResult[3];
    sResult[0]=_J_CurDrive()+'A';
    sResult[1]=':';
    sResult[2]=NIL;
    _retc(sResult);
}

/*Ŀ
** Logical:=J_ISNOVELL()                                                    
**                                                                          
** Return TRUE if Novell, Else return FALSE                                 
**
*****************************************************************************/
CLIPPER J_ISNOVELL(void)
{
    _retl(_J_IsNovell());
}

/*Ŀ
** Logical:=J_ISPROTMODE()                                                  
**                                                                          
** Return TRUE if 'ProtectedMode', Else return FALSE                        
**
*****************************************************************************/
CLIPPER J_ISPROTMO(void)
{
    _retl(_J_IsProtMode());
}

/*Ŀ
** str:=GetTrueName(str) support for J_TRUENAME()                           
**
*****************************************************************************/
HIDE void J_GetTrueName(BYTEP sInpStr,USHORT nInpLen)
{
    BYTEP sOutStr;
    HANDLE xvHandle;
    if(nInpLen>0)
    {
        if(xvHandle=_xvalloc(129,0))
        {
            if((sOutStr=(BYTEP)_xvlock(xvHandle))!=NULL)
            {
                *sOutStr=NIL;
                if(!_J_TrueName(sInpStr,sOutStr)) _retc(sOutStr);
                else (void)J_ErrorHandler(1,EM_TRUENAME);
                _xvunlock(xvHandle);
            }
            else (void)J_ErrorHandler(2,EM_TRUENAME);
            _xvfree(xvHandle);
        }
        else (void)J_ErrorHandler(3,EM_TRUENAME);
    }
    else (void)J_ErrorHandler(4,EM_TRUENAME);
}

/*Ŀ
** String:=J_TRUENAME(String)                                               
** If ERROR then Result==NIL                                                
**                                                                          
** Return TRUE-name of a file                                               
**
*****************************************************************************/
CLIPPER J_TRUENAME(void)
{
    if(PCOUNT==1 && ISCHAR(1)) J_GetTrueName(_parc(1),_parclen(1));
    else (void)J_ErrorHandler(5,EM_TRUENAME);
}

/*Ŀ
** handle:=SemOpen(str) support for J_SEMOPEN()                             
**
*****************************************************************************/
#define SEM_MAX_LEN 127
#define SEM_MAX_VAL 127
HIDE void J_DoSemOpen(BYTEP sInpStr,USHORT nInpLen,USHORT nValue)
{
    BYTEP sSem,sDst;
    ULONG nRet;
    USHORT nLoop;
    HANDLE xvHandle;
    if(nInpLen>0 && nInpLen<=SEM_MAX_LEN && nValue<=SEM_MAX_VAL)
    {
        if(xvHandle=_xvalloc(nInpLen+2,0))
        {
            if((sDst=sSem=(BYTEP)_xvlock(xvHandle))!=NULL)
            {
                sDst++;
                nLoop=nInpLen;
                while(nLoop--) *sDst++=*sInpStr++;
                *sDst=NIL;
                *sSem=nInpLen;
                if(nRet=_J_SemOpen(sSem,nValue,nInpLen+1)) _retnl(nRet);
                else (void)J_ErrorHandler(1,EM_SEMOPEN);
                _xvunlock(xvHandle);
            }
            else (void)J_ErrorHandler(2,EM_SEMOPEN);
            _xvfree(xvHandle);
        }
        else (void)J_ErrorHandler(3,EM_SEMOPEN);
    }
    else (void)J_ErrorHandler(4,EM_SEMOPEN);
}

/*Ŀ
** SemHandle:=J_SEMOPEN("ThisSemaphore"[,Value])                            
** If ERROR then Result==NIL                                                
** If NOT_NOVELL then Result==1                                             
**                                                                          
** Open a semaphore                                                         
**
*****************************************************************************/
CLIPPER J_SEMOPEN(void)
{
    if(PCOUNT==1 && ISCHAR(1)) J_DoSemOpen(_parc(1),_parclen(1),1);
    else
    {
        if(PCOUNT==2 && ISCHAR(1) && ISNUM(2))
          J_DoSemOpen(_parc(1),_parclen(1),_parni(2));
        else (void)J_ErrorHandler(5,EM_SEMOPEN);
    }
}

/*Ŀ
** Value:=J_SEMVALUE(SemHandle)                                             
** If ERROR then Result==NIL                                                
** If NOT_NOVELL then Result==0                                             
**                                                                          
** Val of a semaphore                                                       
**
*****************************************************************************/
CLIPPER J_SEMVALUE(void)
{
    if(PCOUNT==1 && ISNUM(1)) (void)_J_SemVal(_parnl(1));
    else (void)J_ErrorHandler(1,EM_SEMVALUE);
}

/*Ŀ
** Count:=J_SEMCOUNT(SemHandle)                                             
** If ERROR then Result==NIL                                                
** If NOT_NOVELL then Result==0                                             
**                                                                          
** Count of a semaphore                                                     
**
*****************************************************************************/
CLIPPER J_SEMCOUNT(void)
{
    if(PCOUNT==1 && ISNUM(1)) (void)_J_SemOpc(_parnl(1));
    else (void)J_ErrorHandler(1,EM_SEMCOUNT);
}

/*Ŀ
** Logical:=J_SEMWAIT(SemHandle[,nTime])                                    
** If ERROR then Result==NIL                                                
** If NOT_NOVELL then Result==TRUE                                          
**                                                                          
** Wait semaphore                                                           
**
*****************************************************************************/
CLIPPER J_SEMWAIT(void)
{
    if(PCOUNT==1 && ISNUM(1)) _J_SemFn(_parnl(1),0xC502,0);
    else
    {
        if(PCOUNT==2 && ISNUM(1) && ISNUM(2))
          _J_SemFn(_parnl(1),0xC502,_parni(2));
        else (void)J_ErrorHandler(1,EM_SEMWAIT);
    }
}

/*Ŀ
** Logical:=J_SEMSIGNAL(SemHandle)                                          
** If ERROR then Result==NIL                                                
** If NOT_NOVELL then Result==TRUE                                          
**                                                                          
** Signal a semaphore                                                       
**
*****************************************************************************/
CLIPPER J_SEMSIGNA(void)
{
    if(PCOUNT==1 && ISNUM(1)) _J_SemFn(_parnl(1),0xC503,0);
    else (void)J_ErrorHandler(1,EM_SEMSIGNAL);
}

/*Ŀ
** Logical:=J_SEMCLOSE(SemHandle)                                           
** If ERROR then Result==NIL                                                
** If NOT_NOVELL then Result==TRUE                                          
**                                                                          
** Close a semaphore                                                        
**
*****************************************************************************/
CLIPPER J_SEMCLOSE(void)
{
    if(PCOUNT==1 && ISNUM(1)) _J_SemFn(_parnl(1),0xC504,0);
    else (void)J_ErrorHandler(1,EM_SEMCLOSE);
}

